#include <LiquidCrystal.h>
#include <avr/interrupt.h>

#define RS 12
#define EN 11
#define D4 6
#define D5 7
#define D6 8
#define D7 9
#define piezo 13
#define winScore 5

// Pinii asociati
#define p1DownButton 10
#define p1UpButton 5  
#define p2DownButton 2 
#define p2UpButton 3
#define gameStartButton 4  

#define Player1 1
#define Player2 2

// Delay pentru minge
#define updateTimeDiagonal 23
#define updateTimeHorizontal 17

// Pozitia initiala a mingei
#define StartX 35
#define StartY 7

#define Button_Pressed (starButton)

void(* resetFunc) (void) = 0;   // Functia de reset

// Variabile pentru intreruperi
volatile boolean xUp = true;
volatile boolean xDown = true;

LiquidCrystal lcd(RS, EN, D4, D5, D6, D7);    // Crearea obiectului de lcd

// Cand un jucator suteaza
uint8_t scores = 0;

// Dir mingei
uint8_t ballYDir = 0;
uint8_t ballXDir = -1;

// Setam pozitiile initiale
uint8_t ballY = StartY;
uint8_t ballX = StartX;

// Array diplayului unde va fi afisata mingea
uint8_t ballCharArray[8] = {0, 0, 0, 0, 0, 0, 0, 16};

// Paletele
class paddle
{
  public:
  uint8_t paddleMovement[16] = {0,0,0,0,0,4,4,4,0,0,0,0,0,0,0,0};
  
  uint8_t PaddlePos = 6;
  uint8_t Score = 0;

  // Cand apasam "down"
  void MovePaddleDown()
  {
    if(PaddlePos < 14)
    {
      PaddlePos++;
      paddleMovement[PaddlePos-2]=0;
      paddleMovement[PaddlePos+1]=4;
    }
  }

  // Cand apasam "up"
  void MovePaddleUp()
  {
    if(PaddlePos > 1)
    {
      PaddlePos--;
      paddleMovement[PaddlePos+2]=0;
      paddleMovement[PaddlePos-1]=4;
    }
  }

  // Printarea paletelor fieccarui jucator
  void PrintPaddles(uint8_t PlayerNum)
  {
    if(PlayerNum == 2)
    {
      lcd.createChar(0, paddleMovement);
      lcd.createChar(1, paddleMovement+8);

      // Setam caracterul pe poziti 15 de pe linia 1
      lcd.setCursor(14, 0);
      lcd.write(byte(0));

      // Setam caracterul pe poziti 15 de pe linia 1
      lcd.setCursor(14, 1);
      lcd.write(byte(1));
    }
  
    else if(PlayerNum == 1)
    {
      lcd.createChar(2, paddleMovement);
      lcd.createChar(3, paddleMovement+8);
  
      // Setam caracterul pe poziti 2 de pe linia 1
      lcd.setCursor(1, 0);
      lcd.write(byte(2));


      // Setam caracterul pe poziti 2 de pe linia 1
      lcd.setCursor(1, 1);
      lcd.write(byte(3));
    }
  }
};

// Obiecte palete
paddle p1, p2;


// Ce printam cand incepe jocul
void StartGamePrint()
{
  lcd.print(F("   PONG GAME "));
  
  // Setem pozitia butonului
  lcd.setCursor(0, 1);
  lcd.print(F(" PRESS A BUTTON"));

  uint8_t starButton = 0;
  
  // Butonul de start
  while(!(Button_Pressed))
  {

    starButton = ((PIND & (1 << gameStartButton))); 
  }
  lcd.clear();   // Stergem tot ce am afisat
}

void PrintScore()
{
  // Afisam scorul
  lcd.clear();
  lcd.print(F("PLAYER1  PLAYER2"));
  lcd.setCursor(3 ,1);
  lcd.print(p1.Score);
  lcd.setCursor(12 ,1);
  lcd.print(p2.Score);

  delay(1800);

  lcd.clear();
}


void PrintWinner(int PlayerNum)
{
  lcd.setCursor(0 ,0);
  lcd.print(F("   GAME  OVER   "));
  lcd.setCursor(1, 1);
  lcd.print(F("PLAYER "));
  lcd.print(PlayerNum);
  lcd.setCursor(11 ,1);
  lcd.print(F("WINS"));

  // Texta va fi afisat ceva timp si va conta o muzica
  tone(piezo,500,200);
  delay(200);
  tone(piezo,500,200);
  delay(200);
  tone(piezo,500,200);
  delay(200);
  tone(piezo,800,150);
  delay(150);
  tone(piezo,500,500);
  delay(500);
  tone(piezo,600,1000);
  delay(3750);

  // Jocul se reseteaza
  resetFunc();
}

// Declaring member fucntions defined outside of class
void GenerateBallArray()
{
  for(uint8_t i=0; i<8; i++)
    {
    if(i == (ballY % 8))
    {
      ballCharArray[i] = 2 << (4 - (ballX % 5));
    }
    else
    {
      ballCharArray[i] = 0;
    }
  }
}

void PrintBall()
{
  // Calcumunde se va afisa mingea
  uint8_t LCDCol = ballX / 5;

  // Sus sau jos
  uint8_t LCDRow = (ballY <= 7) ? 0 : 1;

  // Cream arrayul
  lcd.createChar(4, ballCharArray);

  // Mutam cursorul
  lcd.setCursor(LCDCol,LCDRow);

  // Desenam caracterul
  lcd.write(byte(4));
}

void UpdateBall(uint8_t P1_PaddlePos, uint8_t P2PaddlePos)
{
  
  // Se face un mic delay cand se reseteza pozitia mingei
  // pe diagonala sau orizontala
  if(ballYDir)     
  {
    delay(updateTimeDiagonal);
  }
  else
  {
    delay(updateTimeHorizontal);
  }
  
  // Daca mingea iese in afara ecranului
  if((ballX <= 6) || (ballX >= 73))
  {
    AwardAPoint();
  }

  // Verificam cand mingea se afla pe partea dreapta a LCD-ului
  else if(ballX == 72)
  {
    // Daca se loveste in mijloc de paleta
    // Schimbam directia si acompaniem cu sunet
    if(ballY == P2PaddlePos)
    {
      ballXDir = -1;
      
            
      tone(13, 1000, 100);
      digitalWrite(A3, HIGH);
      digitalWrite(A3, LOW);
    }

    // Daca se loveste in parte de jos a peletelor
    else if(ballY == (P2PaddlePos + 1))
    {
      ballXDir = -1;
      if(ballY == 15)
      {
        ballYDir = -1;
      }
      else
      {
        ballYDir = 1;
      }
      
            
      tone(13, 1000, 100);
      digitalWrite(A3, HIGH);
      digitalWrite(A3, LOW);
    }

    // Daca se loveste in parte de sus a paletei
    else if(ballY == (P2PaddlePos - 1)){
      ballXDir = -1;
      if(ballY == 0) 
      {
        ballYDir = 1;
      }
      else
      {
        ballYDir = -1;
      }
      
      tone(13, 1000, 100);
      digitalWrite(A3, HIGH);
      digitalWrite(A3, LOW);
    }
  }

  // Daca mingea se afla pe partea stanga a LCD-ului prtocedam la fel
  else if(ballX == 7)
  {
    if(ballY == P1_PaddlePos)
    {
      ballXDir = 1;
      
      tone(13, 1000, 100);
      digitalWrite(A3, HIGH);
      digitalWrite(A3, LOW);
    }

    else if(ballY == (P1_PaddlePos + 1)){
      ballXDir = 1;
      if(ballY == 15)
      {
        ballYDir = -1;
      }
      else
      {
        ballYDir = 1;
      }
      
      tone(13, 1000, 100);
      digitalWrite(A3, HIGH);
      digitalWrite(A3, LOW);
    }

    else if(ballY == (P1_PaddlePos - 1))
    { 
      ballXDir = 1;
      if(ballY == 0)
      {
        ballYDir = 1;
      }
      else
      {
        ballYDir = -1;
      }
      
      tone(13, 1000, 100);
      digitalWrite(A3, HIGH);
      digitalWrite(A3, LOW);
    }
  }

  // Daca se loveste de marginea de sus sau jos a display-ului
  else if((ballY == 0) || (ballY == 15))
  {

    ballYDir *= -1;
    tone(13, 523, 100);
    digitalWrite(A0, HIGH);
    digitalWrite(A0, LOW);
  }
    
  // Resetam dacca vreun jucator marcheaza
  if(scores == 1)
  { 
    ballX = StartX;
    ballY = StartY;
    ballXDir *= -1;
    ballYDir = 0;

    p1.PaddlePos = 6;
    p2.PaddlePos = 6;
    
    for(uint8_t i=0; i<16; i++)
    {
      if((i==5) || (i==6) || (i==7))
      {
        p1.paddleMovement[i] = 4;
        p2.paddleMovement[i] = 4;
      }
      else
      {
        p1.paddleMovement[i] = 0;
        p2.paddleMovement[i] = 0;
      }
    }
    
    scores = 0;
  }

  //  Stergem mingea din locatia curenta
  uint8_t LCDCol = ballX / 5;
  uint8_t LCDRow = (ballY <= 7) ? 0 : 1;
  lcd.setCursor(LCDCol, LCDRow);
  lcd.print(" ");
  
  // Schimbam pozitia
  ballX += ballXDir;
  ballY += ballYDir;
  GenerateBallArray();

  // Afisam mingea dupa recalculare
  PrintBall();
}

void AwardAPoint()
{
  if(ballX <= 8)    // P2 marcheaza
  {
    p2.Score++;
    tone(piezo,400,200);
    delay(200);
    tone(piezo,200,200);
  }
    
  else              // p1 marcheaza canta muzica si se aduna la scor
  {
    p1.Score++;
    tone(piezo,300,200);
    delay(100);
    tone(piezo,150,200);

  }
  
  // Cand cioneva ajunge la 5 puncte
  // Afisam castigatorul 
  
  if(p1.Score == winScore)
  {
    PrintWinner(Player1);
  }

  else if(p2.Score == winScore)
  {
    PrintWinner(Player2);
  }

  PrintScore();
    
  scores = 1;
}


void setup()
{
  // Marimea display-ului
  lcd.begin(16, 2);
  
  // Setari de butoane
  DDRD &= ~(1<<p1UpButton);
  DDRD &= ~(1<<p1DownButton);
  DDRD &= ~(1<<p2UpButton);
  DDRD &= ~(1<<p2DownButton);
  DDRD &= ~(1<<gameStartButton);
  
  // Printam inceputul jocului
  StartGamePrint();
  
  // Printam paletele si mingea
  p1.PrintPaddles(Player1);
  p2.PrintPaddles(Player2);

  PrintBall();

  // Setam intreruperile
  PCMSK2 |= (1 << PCINT21);
  PCICR |= (1 << PCIE2);

  PCMSK0 |= (1 << PCINT2);
  PCICR |= (1 << PCIE0);

  sei();
  
  // Jucatorul 2 va cauza intreruperi
  attachInterrupt(digitalPinToInterrupt(p2DownButton), P2MoveDown, RISING);
  attachInterrupt(digitalPinToInterrupt(p2UpButton), P2MoveUp, RISING);
}

void loop()
{
  // Paletele
  p1.PrintPaddles(Player1);
  p2.PrintPaddles(Player2);
  
  // Mingea pozitie
  UpdateBall(p1.PaddlePos, p2.PaddlePos);
}

void P2MoveDown()
{
  p2.MovePaddleDown();
}

void P2MoveUp()
{
  p2.MovePaddleUp();
}

// Intreruperea butoanelor
ISR(PCINT2_vect)
{
  if(xUp)
  {
    p1.MovePaddleUp();
    xUp = false;
  }
  else
  {
    xUp = true;
  }
}

ISR(PCINT0_vect)
{
  if(xDown)
  {
    p1.MovePaddleDown();
    xDown = false;
  }
  else
  {
    xDown = true;
  }
}